#include <clang-c/Index.h>
#include <string.h>

#include <fstream>
#include <iostream>
#include <string>
#include <vector>

using namespace std;

static constexpr char header_top[] = R"(
#ifndef HEADER_H_H_
#define HEADER_H_H_

struct Foo {
  int bar;
)";
static constexpr char header_bottom[] = R"(
};

#endif  // HEADER_H_H_
)";

static constexpr char cpp_file[] = R"(
#include "header_file.h"

int main() {
  Foo foo;
  foo.bar = 7;
  foo.baz = 8;
}
)";

static constexpr char header_name[] = "/tmp/header_file.h";
static constexpr char cpp_name[] = "/tmp/cpp_file.cpp";

void WriteFile(const string& filename, const string& contents) {
  ofstream ost(filename);
  ost << contents;
}

void DisplayDiagnostics(CXTranslationUnit clang_tu) {
  uint num_diagnostics = clang_getNumDiagnostics(clang_tu);
  if (num_diagnostics == 0) {
    cout << "Hooray! No diagnostics found." << endl;
    return;
  }
  for (uint i = 0; i < num_diagnostics; ++i) {
    auto diag = clang_getDiagnostic(clang_tu, i);
    CXSourceLocation source_location = clang_getDiagnosticLocation(diag);
    cout << clang_getCString(clang_formatDiagnostic(
                diag, clang_defaultDiagnosticDisplayOptions())) << endl;
    clang_disposeDiagnostic(diag);
  }
}

bool ReparseTU(CXTranslationUnit clang_tu, uint num_unsaved_files,
               CXUnsavedFile* unsaved_files) {
  auto failure =
      clang_reparseTranslationUnit(clang_tu, num_unsaved_files, unsaved_files,
                                   clang_defaultReparseOptions(clang_tu));
  if (failure) {
    cerr << "Reparse failed" << endl;
    clang_disposeTranslationUnit(clang_tu);
    return false;
  }
  DisplayDiagnostics(clang_tu);
  return true;
}

int main() {
  // Write initial version of files.
  WriteFile(cpp_name, cpp_file);
  WriteFile(header_name, string(header_top) + header_bottom);

  const auto parse_options = CXTranslationUnit_DetailedPreprocessingRecord |
                             clang_defaultEditingTranslationUnitOptions();
  auto index = clang_createIndex(0, 0);

  cout << "Parsing initial files." << endl;
  auto clang_tu =
      clang_parseTranslationUnit(index, cpp_name, nullptr, 0,  // No flags.
                                 nullptr, 0,  // No unsaved files.
                                 parse_options);
  DisplayDiagnostics(clang_tu);

  cout << endl << "Immediate reparse." << endl;
  if (!ReparseTU(clang_tu, 0, nullptr /* No unsaved files. */)) {
    clang_disposeIndex(index);
    return 1;
  }

  cout << endl << "Parsing with updated header in memory." << endl;
  string new_header_contents =
      string(header_top) + "  int baz;" + header_bottom;
  CXUnsavedFile unsaved_files[] = {
      {header_name, new_header_contents.c_str(), new_header_contents.size()}};
  if (!ReparseTU(clang_tu, 1, unsaved_files)) {
    clang_disposeIndex(index);
    return 1;
  }

  cout << endl << "Saving updated header and reparsing from disk." << endl;
  WriteFile(header_name, new_header_contents);
  if (!ReparseTU(clang_tu, 0, nullptr /* No unsaved files. */)) {
    clang_disposeIndex(index);
    return 1;
  }

  clang_disposeTranslationUnit(clang_tu);
  clang_disposeIndex(index);
}
